import json
from typing import Any

from flask import render_template_string, session
from flask import request as req
from markupsafe import Markup

from ..utils import to_class


class HasRenders:
    def inital_render(self, component_name: str, *args, **kwargs):
        #print(f"Doing initial render {component_name}")
        # if key == "":
        #     key = f"mvlive_{component_name}_{secrets.token_urlsafe()}"
        component_name = component_name + "Component"
        _class: Any = to_class(f"templates.liveflask.{component_name}.{component_name}")
        component_instance = _class()
        component_name: str = _class.__name__

        # print("checking if key is sent via jinja funct")
        key = kwargs.get("key")
        if key:
            # print(f"key found ..... {key}")
            del kwargs["key"]
        else:
            key = getattr(component_instance, "key")
            # print(f"found key {key}")
            # print("Using generated key")
        setattr(component_instance, "key", key)

        # Get props from session
        for prop in component_instance.mvlive__session:
            setattr(
                component_instance,
                prop,
                session.get(
                    prop, getattr(component_instance, prop, None)
                )
            )

        for prop in component_instance.mvlive__url:
            if req.args.get(prop, None):
                setattr(
                    component_instance,
                    prop,
                    req.args.get(prop, None)
                )

        if hasattr(_class, 'boot'):
            component_instance.boot()

        if hasattr(_class, 'mount'):
            component_instance.mount(*args, **kwargs)

        if hasattr(_class, 'booted'):
            component_instance.booted()

        # set_attribute(component_instance, 'key', key)
        html, snapshot = self.to_snapshot(component_instance)
        snapshot_attr: str = json.dumps(snapshot)

        #print(f"Snapshot: {snapshot}")

        return Markup(
            render_template_string(
                """
                    <div data-component="{{ component_name }}" id="{{ key }}" data-snapshot="{{ snapshot_attr }}">
                        {{html|safe}}
                    </div>
                """, snapshot_attr=snapshot_attr, html=html, key=key, component_name=component_name
            )
        )



import hashlib
import json
import pprint
import secrets
from typing import Any

from flask import render_template_string
from markupsafe import Markup

from ..utils import to_class


class HasSnapshots:
    def from_snapshot(self, req_snapshot: dict[str, Any]):
        req_checksum: str = req_snapshot['snapshot']['checksum']
        del req_snapshot['snapshot']['checksum']
        if 'children' in req_snapshot['snapshot']:
            del req_snapshot['snapshot']['children']
        if 'models' in req_snapshot['snapshot']:
            del req_snapshot['snapshot']['models']
        if 'actions' in req_snapshot['snapshot']:
            del req_snapshot['snapshot']['actions']
        if 'polls' in req_snapshot['snapshot']:
            del req_snapshot['snapshot']['polls']
        # pprint.pprint(req_snapshot['snapshot'])

        source_checksum: str = hashlib.md5(
            json.dumps(req_snapshot['snapshot'], sort_keys=True, ensure_ascii=True).encode('utf-8')).hexdigest()

        # TODO: Implement checksum verification
        # if source_checksum != req_checksum:
        #     raise Exception("Stop trying to hack me.")


        class_name: str = req_snapshot['snapshot']['class']
        data: dict[str, Any] = req_snapshot['snapshot']['data']
        # children = req_snapshot['snapshot']['children']

        _class: Any = to_class(f"templates.liveflask.{class_name}.{class_name}")()
        _class.__name__, _class.__class__.__name__ = class_name, class_name
        if getattr(_class, "key", None):
            #print("checking if component key is set.......")
            key = getattr(_class, "key")
            #print(f"found key {key}")
        else:
            key = f"mvlive_{_class.__name__.lower()}_{secrets.token_urlsafe(4)}"
        setattr(_class, "key", key)

        for prop in data.items():
            setattr(_class, prop[0], prop[1])
        return _class

    def to_snapshot(self, _class: Any):
        props: dict[str, Any] = self.get_props(_class)

        # loop through props and if any key matches a key in _class.mvlive__url, add the key value pair to props[url]
        if hasattr(_class, "mvlive__url"):
            props["url"] = {}
            for key in _class.mvlive__url:
                props["url"].update({key: props[key]})



        _class.key = props.get("key")

        # remove all keys from props that begin with '_' or '__'
        # for key in list(props.keys()):
        #     if key.startswith("_"):
        #         del props[key]

        #print(props)
        if _class.render.__doc__ is None:
            html: str = _class.render(
                props | {"this": Markup(f"document.getElementById('{ _class.key }').__liveflask")}
            )
        else:
            html: str = render_template_string(
                _class.render.__doc__,
                **props
            )

        # meta = self.dehydrate_properties(props)

        key = getattr(_class, "key", "")

        if "relationship_result_set" in props:
            del props["relationship_result_set"]

        if "result_set" in props:
            del props["result_set"]



        # loop through props and remove any methods
        for key in list(props.keys()):
            if callable(props[key]):
                #print(key)
                del props[key]


        # loop through props and ensure that all values are cast to original types
        for key in list(props.keys()):
            if isinstance(props[key], str):
                try:
                    props[key] = json.loads(props[key])
                except json.JSONDecodeError:
                    pass




        snapshot: dict[str, Any] = {
            "class": _class.__class__.__name__,
            "key": key,
            "data": props,
            # "html": html,
        }

        snapshot['checksum']: str = hashlib.md5(
            json.dumps(snapshot, sort_keys=True, ensure_ascii=True).encode('utf-8')).hexdigest()

        return html, snapshot



import functools


def url(arg: list | None = None):
    def decorator(cls):
        @functools.wraps(cls, updated=())
        class DecoratedClass(cls):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.__class__.__name__ = cls.__name__

        cls.mvlive__url = []

        if arg is not None:
            for query_param in arg:
                # Assuming Request.input() does something meaningful
                # Here, we attach the result of Request.input() to the class
                cls.mvlive__url.append(query_param)
        return DecoratedClass

    return decorator


def locked(arg: list | None = None):
    def decorator(cls):
        @functools.wraps(cls, updated=())
        class DecoratedClass(cls):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.__class__.__name__ = cls.__name__

        cls.mvlive__locked = []

        if arg is not None:
            for locked_prop in arg:
                # Assuming Request.input() does something meaningful
                # Here, we attach the result of Request.input() to the class
                cls.mvlive__locked.append(locked_prop)
        return DecoratedClass

    return decorator


def session(arg: list | None = None):
    def decorator(cls):
        @functools.wraps(cls, updated=())
        class DecoratedClass(cls):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.__class__.__name__ = cls.__name__

        cls.mvlive__session = []

        if arg is not None:
            for session_binded_prop in arg:
                # Assuming Request.input() does something meaningful
                # Here, we attach the result of Request.input() to the class
                cls.mvlive__session.append(session_binded_prop)
        return DecoratedClass

    return decorator



from typing import Any, NoReturn
from masoniteorm.collection import Collection


class HasProps:
    def get_props(self, component):
        props: dict[str, Any] = dict()
        method_list = [attribute for attribute in dir(component) if
                       callable(getattr(component, attribute)) and attribute.startswith('__') is False]

        for key in component.__dict__.items():
            #print(key)
            if key[0].startswith("_"):
                continue
            if key[0] in method_list:
                continue
            else:
                props[f'{key[0]}'] = key[1]
        for method in method_list:
            if not method.startswith("updated_") and not method.startswith("updating_"):
                props[f'{method}'] = getattr(component, method)
            else:
                pass

        #print(props)
        return props

    def set_props(self, _class, prop, val) -> NoReturn:
        if prop in _class.mvlive__locked:
            raise AttributeError(f"Property [{prop}] is locked from client-side editing.")
        # generate names for hook methods using prop name
        updated_hook: str = f'updated_{prop}'
        updating_hook: str = f'updating_{prop}'

        if hasattr(_class, 'updating'):
            _class.updating(prop, val)

        if hasattr(_class, updating_hook):
            getattr(_class, updating_hook)(val)

        if "." in prop:
            self.access_nested_dict_in_component(_class, prop, val)
        else:
            setattr(_class, prop, val)


        new_val = getattr(_class, prop)
        if hasattr(_class, 'updated'):
            _class.updated(prop, new_val)

        if hasattr(_class, updated_hook):
            getattr(_class, updated_hook)(new_val)

        #Session().set(prop, new_val)



    def dehydrate_props(self, props):
        data = {}
        meta = {}

        for key in props.items():
            if isinstance(key[1], Collection):
                value = key[1].to_dict()
                meta[key[0]] = 'collection'
                data[key[1]] = value
            else:
                pass
        return data, meta

    def access_nested_dict_in_component(self, _class: Any, prop: str, val: str):
        props: dict[str, Any] = self.get_props(_class)
        for key in props.keys():
            if key == prop.split('.')[0]:
                _class.__dict__[key][prop.split('.')[1]] = val



from pydoc import locate
from typing import Any, List, AnyStr


def to_class(path: str) -> object | None:
    """
    Converts string class path to a Python class.

    Args:
        path (str): The string representing the class path.

    Returns:
        Union[type, None]: The Python class if found, otherwise None.
    """
    try:
        class_instance = locate(path)
    except ImportError:
        print('Module does not exist')
        return None
    return class_instance


def set_attribute(obj: Any, path_string: str, new_value: Any):
    parts: List[AnyStr] = path_string.split('.')
    final_attribute_index: int = len(parts) - 1
    current_attribute = obj
    i: int = 0
    for part in parts:
        new_attr: Any | None = getattr(current_attribute, part, None)
        if current_attribute is None:
            print('Error %s not found in %s' % (part, current_attribute))
            break
        if i == final_attribute_index:
            setattr(current_attribute, part, new_value)
        current_attribute: Any = new_attr
        i += 1


def get_attribute(obj, path_string):
    parts: List[Any] = path_string.split('.')
    final_attribute_index: int = len(parts) - 1
    current_attribute: Any = obj
    i: int = 0
    for part in parts:
        new_attr: Any | None = getattr(current_attribute, part, None)
        if current_attribute is None:
            print('Error %s not found in %s' % (part, current_attribute))
            return None
        if i == final_attribute_index:
            return getattr(current_attribute, part)
        current_attribute: Any = new_attr
        i += 1


def dict_diff_changed_values(dict1, dict2):
    # Find keys that are in both dicts but have different values
    keys_changed = {key for key in dict1.keys() & dict2.keys() if dict1[key] != dict2[key]}

    # Create a new dictionary to store the new values
    diff = {}
    for key in keys_changed:
        diff[key] = dict2[key]

    return diff